/* adrian 07jul2004: 
 *	DspOutOss::writeBuffer() fixed bug with stereo mode. Does now work
 *	if popup AudioPreferences|OSS Device Mode is ReadOnly/WriteOnly
 *
 * adrian 14jul2004: call Acoustic Echo Cancellation in stereo_to_mono()
 * adrian 24aug2004: UpdateThreshold from -50dB to -30dB
 * adrian 11oct2004: NLMS_LEN is 240ms, NLPAttenuation is 0dB (there is no AES)
 */			

/***************************************************************
A.1 aec.h
***************************************************************/

#ifndef _AEC_H	/* include only once */

/* aec.h
 * Acoustic Echo Cancellation NLMS-pw algorithm
 * Author: Andre Adrian, DFS Deutsche Flugsicherung
 * <Andre.Adrian@dfs.de>
 *
 * Version 1.1
 */

/* dB Values */
const float M0dB = 1.0f;
const float M3dB = 0.71f;
const float M6dB = 0.50f;
 
/* dB values for 16bit PCM */
const float M10dB_PCM = 10362.0f;   
const float M20dB_PCM = 3277.0f;
const float M25dB_PCM = 1843.0f;
const float M30dB_PCM = 1026.0f;
const float M35dB_PCM = 583.0f;
const float M40dB_PCM = 328.0f;   
const float M45dB_PCM = 184.0f;
const float M50dB_PCM = 104.0f;
const float M55dB_PCM = 58.0f;
const float M60dB_PCM = 33.0f;

const float MAXPCM = 32767.0f;

/* Design constants (Change to fine tune the algorithms */

/* For Normalized Least Means Square - Pre-whitening */
#define NLMS_LEN  (240*8)      	  /* maximum NLMS filter length in taps */
const float PreWhiteAlphaTF = (4000.0f/8000.0f);   /* FIR controls Transfer Frequency */

/* for Geigel Double Talk Detector */
const float GeigelThreshold = M3dB;
const int Thold = 30*8;       	      	  /* DTD hangover in taps */
const float UpdateThreshold = M30dB_PCM;

/* for Non Linear Processor */
const float NLPAttenuation = M0dB;

/* Below this line there are no more design constants */


/* Exponential Smoothing or IIR Infinite Impulse Response Filter */
class IIR_HP {
  float lowpassf;
  float alphaTF;  /* controls Transfer Frequency */

public:
  IIR_HP() {
    lowpassf = 0.0f;
    alphaTF = 0.0f;
  }
  
  void init(float alphaTF_) {
    alphaTF = alphaTF_;
  }

  float highpass(float in) {
    /* Highpass = Signal - Lowpass. Lowpass = Exponential Smoothing */
    lowpassf += alphaTF*(in - lowpassf);
    return in - lowpassf;
  }
};


#define POL   	6     	/* -6dB attenuation per octave per Pol */

class IIR_HP6 {
  float lowpassf[2*POL+1];
  float highpassf[2*POL+1];

public:
  IIR_HP6();
  float highpass(float in) {
    const float AlphaHp6 = 0.075; /* controls Transfer Frequency */
    const float Gain6   = 1.45f;  /* gain to undo filter attenuation */

    highpassf[0] = in;
    int i;
    for (i = 0; i < 2*POL; ++i) {
      /* Highpass = Signal - Lowpass. Lowpass = Exponential Smoothing */
      lowpassf[i+1] += AlphaHp6*(highpassf[i] - lowpassf[i+1]);
      highpassf[i+1] = highpassf[i] - lowpassf[i+1];
    }
    return Gain6*highpassf[2*POL];   
  }
};

 /* Recursive single pole FIR Finite Impulse response filter */
class FIR1 {
  float a0, a1, b1;
  float last_in, last_out;
  
public:
  FIR1();
	void init(float preWhiteTransferAlpha);
  float highpass(float in)  {
    float out = a0 * in + a1 * last_in + b1 * last_out;
    last_in = in;
    last_out = out;

    return out;
  }
};

 
#define NLMS_EXT  (10*8)    // Extention in taps to reduce mem copies
#define DTD_LEN 16          // block size in taps to optimize DTD calculation


class AEC {
  // Time domain Filters
  IIR_HP6 hp0;              // 300Hz cut-off Highpass
  IIR_HP hp1;               // DC-level remove Highpass)
  FIR1 Fx, Fe;              // pre-whitening Highpass for x, e
    
  // Geigel DTD (Double Talk Detector)
  float max_max_x;                // max(|x[0]|, .. |x[L-1]|)
  int hangover;
  float max_x[NLMS_LEN/DTD_LEN];  // optimize: less calculations for max()
  int dtdCnt;                     
  int dtdNdx;
  
  // NLMS-pw
  float x[NLMS_LEN+NLMS_EXT]; 	  // tap delayed loudspeaker signal
  float xf[NLMS_LEN+NLMS_EXT];    // pre-whitening tap delayed signal
  float w[NLMS_LEN];	      	    // tap weights
  int j;      	      	      	  // optimize: less memory copies
  int lastupdate;     	      	  // optimize: iterative dotp(x,x)
  double dotp_xf_xf;              // double to avoid loss of precision
  
public:
  AEC();
  
/* Geigel Double-Talk Detector
 *
 * in d: microphone sample (PCM as floating point value)
 * in x: loudspeaker sample (PCM as floating point value)
 * return: 0 for no talking, 1 for talking
 */
  int dtd(float d, float x);
  
/* Normalized Least Mean Square Algorithm pre-whitening (NLMS-pw)
 * The LMS algorithm was developed by Bernard Widrow
 * book: Widrow/Stearns, Adaptive Signal Processing, Prentice-Hall, 1985
 *
 * in mic: microphone sample (PCM as floating point value)
 * in spk: loudspeaker sample (PCM as floating point value)
 * in update: 0 for convolve only, 1 for convolve and update 
 * return: echo cancelled microphone sample
 */
  float nlms_pw(float mic, float spk, int update);
  
/* Acoustic Echo Cancellation and Suppression of one sample
 * in   d:  microphone signal with echo
 * in   x:  loudspeaker signal
 * return:  echo cancelled microphone signal
 */
  int AEC::doAEC(int d, int x);
};

#define _AEC_H
#endif


/***************************************************************
A.2 aec.cxx
***************************************************************/
/* aec.cxx
 * Acoustic Echo Cancellation NLMS-pw algorithm
 * Author: Andre Adrian, DFS Deutsche Flugsicherung
 * <Andre.Adrian@dfs.de>
 *
 * Version 1.1
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
// #include "aec.h"


IIR_HP6::IIR_HP6()
{
  memset(this, 0, sizeof(IIR_HP6));
}


/* Vector Dot Product */
float dotp(float a[], float b[]) {
  float sum0 = 0.0, sum1 = 0.0;
  int j;
  
  for (j = 0; j < NLMS_LEN; j+= 2) {
    // optimize: partial loop unrolling
    sum0 += a[j] * b[j];
    sum1 += a[j+1] * b[j+1];
  }
  return sum0+sum1;
}


/*
 * Algorithm:  Recursive single pole FIR high-pass filter
 *
 * Reference: The Scientist and Engineer's Guide to Digital Processing
 */

FIR1::FIR1()
{
}

void FIR1::init(float preWhiteTransferAlpha)
{
  float x = exp(-2.0 * M_PI * preWhiteTransferAlpha);
  
  a0 = (1.0f + x) / 2.0f;
  a1 = -(1.0f + x) / 2.0f;
  b1 = x;
  last_in = 0.0f;
  last_out = 0.0f;
}

AEC::AEC()
{
  hp1.init(0.01f);  /* 10Hz */
  Fx.init(PreWhiteAlphaTF);
  Fe.init(PreWhiteAlphaTF);
  
  max_max_x = 0.0f;
  hangover = 0;
  memset(max_x, 0, sizeof(max_x));
  dtdCnt = dtdNdx = 0;
  
  memset(x, 0, sizeof(x));
  memset(xf, 0, sizeof(xf));
  memset(w, 0, sizeof(w));
  j = NLMS_EXT;
  lastupdate = 0;
  dotp_xf_xf = 0.0f;
}


float AEC::nlms_pw(float mic, float spk, int update)
{
  float d = mic;      	        // desired signal
  x[j] = spk;
  xf[j] = Fx.highpass(spk);     // pre-whitening of x

  // calculate error value (mic signal - estimated mic signal from spk signal)
  float e = d - dotp(w, x + j);
  float ef = Fe.highpass(e);    // pre-whitening of e
  if (update) {    
    if (lastupdate) {
      // optimize: iterative dotp(xf, xf)
      dotp_xf_xf += (xf[j]*xf[j] - xf[j+NLMS_LEN-1]*xf[j+NLMS_LEN-1]);
    } else {
      dotp_xf_xf = dotp(xf+j, xf+j);
    } 
      
    // calculate variable step size
    float mikro_ef = 0.5f * ef / dotp_xf_xf;
    
    // update tap weights (filter learning)
    int i;
    for (i = 0; i < NLMS_LEN; i += 2) {
      // optimize: partial loop unrolling
      w[i] += mikro_ef*xf[i+j];
      w[i+1] += mikro_ef*xf[i+j+1];
    }
  }
  lastupdate = update;
  
  if (--j < 0) {
    // optimize: decrease number of memory copies
    j = NLMS_EXT;
    memmove(x+j+1, x, (NLMS_LEN-1)*sizeof(float));    
    memmove(xf+j+1, xf, (NLMS_LEN-1)*sizeof(float));    
  }
  
  return e;
}


int AEC::dtd(float d, float x)
{
  // optimized implementation of max(|x[0]|, |x[1]|, .., |x[L-1]|):
  // calculate max of block (DTD_LEN values)
  x = fabsf(x);
  if (x > max_x[dtdNdx]) {
    max_x[dtdNdx] = x;
    if (x > max_max_x) {
      max_max_x = x;
    }
  }
  if (++dtdCnt >= DTD_LEN) {
    dtdCnt = 0;
    // calculate max of max
    max_max_x = 0.0f;
    for (int i = 0; i < NLMS_LEN/DTD_LEN; ++i) {
      if (max_x[i] > max_max_x) {
        max_max_x = max_x[i];
      }
    }
    // rotate Ndx
    if (++dtdNdx >= NLMS_LEN/DTD_LEN) dtdNdx = 0;
    max_x[dtdNdx] = 0.0f;
  }

  // The Geigel DTD algorithm with Hangover timer Thold
  if (fabsf(d) >= GeigelThreshold * max_max_x) {
    hangover = Thold;
  }
    
  if (hangover) --hangover;
  
  if (max_max_x < UpdateThreshold) {
    // avoid update with silence or noise
    return 1;
  } else {
    return (hangover > 0);
  }
}


int AEC::doAEC(int d, int x) 
{
  float s0 = (float)d;
  float s1 = (float)x;
  
  // Mic Highpass Filter - telephone users are used to 300Hz cut-off
  s0 = hp0.highpass(s0);

  // Spk Highpass Filter - to remove DC
  s1 = hp1.highpass(s1);

  // Double Talk Detector
  int update = !dtd(s0, s1);

  // Acoustic Echo Cancellation
  s0 = nlms_pw(s0, s1, update);

  // Acoustic Echo Suppression
  if (update) {
    // Non Linear Processor (NLP): attenuate low volumes
    s0 *= NLPAttenuation;
  }
  
  // Saturation
  if (s0 > MAXPCM) {
    return (int)MAXPCM;
  } else if (s0 < -MAXPCM) {
    return (int)-MAXPCM;
  } else {
    return (int)roundf(s0);
  }
}


AEC aec;

/* ================================================================ */

#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <qobject.h>

#include "dspoutoss.h"

#include "../callcenterVars.h"

DspOutOss::DspOutOss( const QString &fileName )
	: DspOut()
{
	filename = fileName;
	setDeviceName( "oss" );
	audio_fd = -1;
	if( echoCancel )
		audio_stereo = true; // use stereo for AEC
	else
		audio_stereo = false; // without AEC use standard: mono
}

static int myfd[2] = {-1, -1};
  
// adrian 15jul2004: check open/close of /dev/dsp
int myopen(const char *pathname, int flags)
{
	int i = flags & 1;
	
	if (myfd[i] >= 0) {
		printf( "myopen close %s %d\n", pathname, flags );
		// close( myfd[i] );
	} else {
		printf( "myopen %s %d\n", pathname, flags );
	}
	myfd[i] = open( pathname, flags );
	return myfd[i];
}

// adrian 16jul2004: check open/close of /dev/dsp
int myclose(int fd)
{
	printf( "myclose %d\n", fd );
	int ret = close(fd);
	int i;
	for (i = 0; i < 2; ++i) {
		if (fd == myfd[i]) {
			myfd[i] = -1;
		}
	}
	return ret;
}

DspOutOss::~DspOutOss( void )
{
	myclose( audio_fd );
}

bool DspOutOss::openDevice( DeviceMode mode )
{
	printf("DspOutOss::openDevice\n");

	int oflag;
	if( mode == ReadOnly ) {
		oflag = O_RDONLY;
	} else if( mode == WriteOnly ) {
		oflag = O_WRONLY;
	} else {
		oflag = O_RDWR;
	}
	if( devstate == DeviceOpened ) {
		lasterror = "Device Already Open";
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	// Open immediately or else
	audio_fd = myopen( filename.ascii(), oflag | O_NONBLOCK );
	if( audio_fd == -1 ) {
		lasterror = "Open Failed";
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	// Remove O_NONBLOCK
	int flags = fcntl( audio_fd, F_GETFL );
	flags &= ~O_NONBLOCK;
	fcntl( audio_fd, F_SETFL, flags );

	// keep fragsize less than 20ms !!
	int frag = ( ( 32767 << 16 ) | 7 );
	if( ioctl( audio_fd, SNDCTL_DSP_SETFRAGMENT, &frag ) ) {
		lasterror = QString( "SETFRAG" ) + QString( strerror( errno ) );
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	int format = AFMT_S16_LE;

	if( ioctl( audio_fd, SNDCTL_DSP_SETFMT, &format ) == -1 ) {
		lasterror = QString( "SETFMT" ) + QString( strerror( errno ) );
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	if( format != AFMT_S16_LE ) {
		lasterror = "Format not supported";
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	int channels;
	if( echoCancel ) {
		channels = 2; // adrian use stereo for AEC
	}
	else {
		channels = 1; // without AUC using standard: mono
	}

	if( ioctl( audio_fd, SNDCTL_DSP_CHANNELS, &channels ) == -1 ) {
		lasterror = QString( "DSP_STEREO" ) + QString( strerror( errno ) );
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	if( channels != 1 && channels != 2 ) {
		lasterror = QObject::tr("Unsupported Number of Channels");
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	if( channels == 2 ) {
		printf( QObject::tr("WARNING: nonmonophonic driver, forced into stereo mode (experimental)") + "\n" );
		audio_stereo=true;
	} else {
		printf( QObject::tr("WARNING: monophonic driver") + "\n" );
		audio_stereo=false;
	}

	rate = 8000;

	if( ioctl( audio_fd, SNDCTL_DSP_SPEED, &rate ) == -1 ) {
		lasterror = QString( "DSP_SPEED" ) + QString( strerror( errno ) );
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	if( rate != 8000 ) {
		printf( QObject::tr("WARNING: driver rounded %d Hz request to %d Hz, off by %f%%") + "\n",
			8000, rate, 100*((rate-8000)/8000.0) );
	}


	audio_buf_info info;

	if( mode == WriteOnly ) {
		if( ioctl( audio_fd, SNDCTL_DSP_GETOSPACE, &info ) == -1 ) {
			lasterror = QString( "GETISPACE" ) + QString( strerror( errno ) );
			printf( "ERROR: %s\n", lasterror.ascii() );
			return false;
		}
	} else {
		if( ioctl( audio_fd, SNDCTL_DSP_GETISPACE, &info ) == -1 ) {
			lasterror = QString( "GETOSPACE" ) + QString( strerror( errno ) );
			printf( "ERROR: %s\n", lasterror.ascii() );
			return false;
		}
	}
	audio_buf.resize( info.fragsize * sizeof( short ) );

	lasterror = QString::null;
	devstate = DeviceOpened;
	return true;

}

bool DspOutOss::openDevice( int exist_fs )
{
	audio_fd = exist_fs;
	if( audio_fd == -1 ) {
		lasterror = "Open Failed";
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	audio_buf_info info;
	if( ioctl( audio_fd, SNDCTL_DSP_GETISPACE, &info ) == -1 ) {
		lasterror = QString( "GETISPACE" ) + QString( strerror( errno ) );
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}
	audio_buf.resize( info.fragsize * sizeof( short ) );

	lasterror = QString::null;
	devstate = DeviceOpened;
	return true;
}



// count is in 16-bit words in the smaller of the two buffers
void DspOutOss::mono_to_stereo(void *b_from, void *b_to, size_t count)
{
printf( "mono_to_stereo\n" );//!!!
	int changed = 0;
	double max;
	scale_mono_input(b_from, count, speakerVol, changed, max);
	if( changed )
		speakerVolUpdated();
	speakerSigUpdated( max );
	int8_t *b1 = reinterpret_cast<int8_t*>(b_from);
	int8_t *b2 = reinterpret_cast<int8_t*>(b_to);
	for( size_t i=0; i < count; i++ ) {
		int8_t x0 = b1[2*i];
		int8_t x1 = b1[2*i+1];
		b2[4*i]   = x0;
		b2[4*i+1] = x1;
		b2[4*i+2] = x0;
		b2[4*i+3] = x1;
	}
}

// count is in 16-bit words in the smaller of the two buffers
void DspOutOss::stereo_to_mono(void *b_from, void *b_to, size_t count)
{
/*
echoCancel is 1 every time this function is called because audio_stereo is 1
	if( !echoCancel ) { //!
		int *b1 = reinterpret_cast<int*>(b_from);
		int *b2 = reinterpret_cast<int*>(b_to);
		for( size_t i=0; i < count; i++ ) {
			b2[i] = (int)(microVol * b1[2*i]);
		}
	}
	else {
*/
	// adrian use acoustic echo cancellation
	short *b1 = reinterpret_cast<short*>(b_from);
	short *b2 = reinterpret_cast<short*>(b_to);
	for( size_t i=0; i < count; i++ ) {
		b2[i] = aec.doAEC(b1[2*i], b1[2*i+1]);
	}
	int changed = 0;
	double max;
	scale_mono_input(b2, count, microVol, changed, max);
	if( changed )
		microVolUpdated();
	microSigUpdated( max );
}

// count is in 16-bit words in the smaller of the two buffers
void DspOutOss::scale_mono_input(void *b_to, size_t count, double& vol, int& changed, double& maxvol)
{
	short *b2 = reinterpret_cast<short*>(b_to);
	int max = 0;
	for( size_t i=0; i < count; i++ ) {
		if( max < abs( (int)b2[i] ) ) {
			max = abs( (int)b2[i] );
		}
	}
	maxvol = vol * max;
	if( maxvol > MAXPCM ) {
		changed = 1;
		vol = MAXPCM / max;
		maxvol = MAXPCM;
	}
	else
		changed = 0;
	for( size_t i=0; i < count; i++ ) {
		b2[i] = (short)(vol * b2[i]);
	}
}

bool DspOutOss::writeBuffer( void )
{
	if( devstate != DeviceOpened ) {
		lasterror = "Device Not Open";
		return false;
	}
	size_t count = audio_buf.getSize();
	void *buf;

	int channels;
	if( echoCancel )
		channels = 2;
	else
		channels = 1;
	if( ioctl( audio_fd, SNDCTL_DSP_CHANNELS, &channels ) == -1 ) {
		lasterror = QString( "DSP_STEREO" ) + QString( strerror( errno ) );
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}
	if( channels == 2 )
		audio_stereo=true;
	else
		audio_stereo=false;

	if( audio_stereo ) {
		buf = alloca(2*count);
		mono_to_stereo(audio_buf.getData(), buf, count/2);
		count *= 2;
	} else {
		buf = audio_buf.getData();
		int changed = 0;
		double max;
		scale_mono_input(buf, count/2, speakerVol, changed, max);
		if( changed )
			speakerVolUpdated();
		speakerSigUpdated( max );
	}
	audio_buf_info info;
	if( ioctl( audio_fd, SNDCTL_DSP_GETOSPACE, &info ) == 0 ) {
		if( info.fragstotal - info.fragments > 15 ) {
			// drop the fragment if the buffer starts to fill up
			return true;
		}
	}
	for(;;) {
		// if( write( audio_fd, audio_buf.getData(), audio_buf.getSize() ) != -1 ) {
		// adrian bug fix
		if( write( audio_fd, buf, count ) != -1 ) {
			break;
		}
	}

	return true;
}

unsigned int DspOutOss::readableBytes( void )
{
	audio_buf_info info;
	struct timeval timeout;
	fd_set read_fds;

	if( devstate != DeviceOpened ) {
		return 0;
	}
	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
	FD_ZERO( &read_fds );
	FD_SET( audio_fd, &read_fds );
	if( select( audio_fd + 1, &read_fds, NULL, NULL, &timeout ) == -1 ) {
		return 0;
	}
	if( !FD_ISSET( audio_fd, &read_fds ) ) {
		return 0;
	}
	if( ioctl( audio_fd, SNDCTL_DSP_GETISPACE, &info ) == -1 ) {
		lasterror = strerror( errno );
		return 0;
	}
	return info.bytes;
}


bool DspOutOss::readBuffer( int bytes )
{
	if( devstate != DeviceOpened ) {
		lasterror = "Device Not Open";
		return false;
	}

	int channels;
	if( echoCancel )
		channels = 2;
	else
		channels = 1;
	if( ioctl( audio_fd, SNDCTL_DSP_CHANNELS, &channels ) == -1 ) {
		lasterror = QString( "DSP_STEREO" ) + QString( strerror( errno ) );
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}
	if( channels == 2 )
		audio_stereo=true;
	else
		audio_stereo=false;

	audio_buf.resize( bytes );
	size_t count = bytes;

	void *buf;
	if( audio_stereo ) {
		count *= 2;
		buf = alloca(count);
	} else {
		buf = audio_buf.getData();
	}

	ssize_t rc = read( audio_fd, buf, count );
	if( rc < 0 ) {
		lasterror = strerror(errno);
		return false;
	}
	if( (size_t)rc != count ) {
		/* should we treat rc == 0 (EOF) specially? */
		printf("warning: asked microphone for %lu bytes but got %lu\n",
			(unsigned long)count, (unsigned long)rc);
		/* assert: rc < count */
		/* Does count have to be a multiple of 2 or even 4? */
		count = rc;
		audio_buf.resize( audio_stereo? count/2 : count );
	}

	if( audio_stereo ) {
		stereo_to_mono(buf, audio_buf.getData(), count/4);
	}
	else {
		int changed = 0;
		double max;
		scale_mono_input(audio_buf.getData(), count/2, microVol, changed, max);
		if( changed )
			microVolUpdated();
		microSigUpdated( max );
	}

	return true;
}
